#Dataset Settings
gebco_path = '/home/vres-climate/data/GEBCO_2014_2D.nc'
cutout_dir = '/home/vres/data/cutouts'
ncep_dir = '/home/vres-climate/data/rda_ucar'
cordex_dir = '/home/vres-climate/data/cordex/RCP8.5'
sarah_dir = '/home/vres-climate/data/sarah_v2'
weather_dataset = {'module': 'cordex', 'model': 'MPI-M-MPI-ESM-LR'}
